package de.kesystem.remotemessadapter;

import android.app.AlertDialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.DialogInterface.OnClickListener;
import android.os.Environment;
import android.text.Editable;
import android.view.Gravity;
import android.view.View;
import android.view.ViewGroup;
import android.view.ViewGroup.LayoutParams;
import android.widget.AdapterView;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.LinearLayout;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

public class KEFileDialog {
    public String StartDirectory = "";
    public String DefaultFilename = "";
    public String FileExtFilter = "*";
    private int FileOpen     = 0;          //Die möglichen Arten des Dialogs
    private int FileSave     = 1;
    private int FolderChoose = 2;
    private int DialogType = FileSave;
    private Context ThisContext;
    private TextView TitleText;            //"Lade .xxx aus /..." oder "Sichere .xxx nach /..."
    private ListView FileDirListView;              //Liste der Verzeichnisse und Files
    private EditText FileInputText;             //Eingabefeld fuer Filename
    private String SelectedFileDir = DefaultFilename;
    private String AktDirectory = "";
    private List<String> FileDirList = null;
    private KEFileDialogInterface KEFileDialogListener = null;
    private ArrayAdapter<String> FileDirListAdapter = null;


    // Callback interface für File/Directory-Selection
    public interface KEFileDialogInterface {
        public void onOKClick(String SelFileDir);
        public void onCancelClick();
    }

    //Konstruktor
    public KEFileDialog(Context context, String dialog_type, KEFileDialogInterface KEFileDialogInterface) {
        if (dialog_type.equals("FileOpen"))          DialogType = FileOpen;
        else if (dialog_type.equals("FileSave"))     DialogType = FileSave;
        else if (dialog_type.equals("FolderChoose")) DialogType = FolderChoose;
        else DialogType = FileOpen;

        ThisContext = context;
        if(StartDirectory=="") StartDirectory = Environment.getExternalStorageDirectory().getAbsolutePath();
        KEFileDialogListener = KEFileDialogInterface;

        try {
            StartDirectory = new File(StartDirectory).getCanonicalPath();
        } catch (IOException ioe) {
        }
    }

    //Start des FileDialogs, wenn kein Start-Verzeichnis angegeben ist
    public void Generate() {
        if (AktDirectory.equals("")) Generate(StartDirectory);
        else Generate(AktDirectory);
    }

    //Start des FileDialogs, wenn ein Start-Verzeichnis angegeben ist. Erzeugung des Dialogs mit allem drum und dran
    public void Generate(String dir) {
        //File/Directory-Liste erzeugen
        File dirFile = new File(dir);
        if (! dirFile.exists() || ! dirFile.isDirectory()) dir = StartDirectory; //Wenn Müll übergeben wurde: Im Standard-Startverzeichnis beginnen
        try {
            dir = new File(dir).getCanonicalPath();
        } catch (IOException ioe) {
            return;
        }
        AktDirectory = dir;
        FileDirList = getFilesAndDirs(dir);

        //Layout für Titelzeile und "NewFolder"-Button
        LinearLayout TitleLayout = new LinearLayout(ThisContext);
        TitleLayout.setOrientation(LinearLayout.VERTICAL);
        //Titelzeile
        TitleText = new TextView(ThisContext);
        TitleText.setLayoutParams(new LayoutParams(LayoutParams.MATCH_PARENT, LayoutParams.WRAP_CONTENT));
        if (DialogType == FileOpen    ) TitleText.setText("Lade " + FileExtFilter + " aus " + dir);
        if (DialogType == FileSave    ) TitleText.setText("Sichere " + FileExtFilter + " nach " + dir);
        if (DialogType == FolderChoose) TitleText.setText("Folder Select:");
        TitleText.setGravity(Gravity.CENTER_VERTICAL);
        TitleText.setBackgroundColor(-12303292); // dark gray 	-12303292
        TitleText.setTextColor(ThisContext.getResources().getColor(android.R.color.white) );
        TitleLayout.addView(TitleText);
        //New Folder-Button im Falle von Save-Dialog oder FolderChose
        if (DialogType == FolderChoose || DialogType == FileSave) {
            Button newDirButton = new Button(ThisContext);
            newDirButton.setLayoutParams(new LayoutParams(LayoutParams.MATCH_PARENT, LayoutParams.WRAP_CONTENT));
            newDirButton.setText("Neues Verzeichnis");
            newDirButton.setOnClickListener(new View.OnClickListener() {
                @Override
                public void onClick(View v) {
                    final EditText input = new EditText(ThisContext);
                    //New Folder ist ein eigener Dialog im übergeordneten FileDialog
                    new AlertDialog.Builder(ThisContext).setTitle("Name des neuen Verzeichnisses").setView(input).setPositiveButton("OK", new OnClickListener() {
                        public void onClick(DialogInterface dialog, int whichButton) {
                            Editable newDir = input.getText();
                            String newDirName = newDir.toString();
                            //Neues Verz. erstellen und direkt reinnavigieren
                            if ( CreateDirectory(AktDirectory + "/" + newDirName) ) {
                                AktDirectory += "/" + newDirName;
                                UpdateDialogViews();
                            } else {
                                Toast.makeText(ThisContext, "Fehler beim Erstellen von Verzeichnis '" + newDirName, Toast.LENGTH_SHORT).show();
                            }
                        }
                    }).setNegativeButton("Cancel", null).show();
                }
            });
            TitleLayout.addView(newDirButton);
        }

        //Layout für Texteingabe und File-/Directory-Liste
        LinearLayout InputLayout = new LinearLayout(ThisContext);
        InputLayout.setOrientation(LinearLayout.VERTICAL);
        if (DialogType == FileOpen || DialogType == FileSave) {
            FileInputText = new EditText(ThisContext);
            FileInputText.setText(DefaultFilename);
            InputLayout.addView(FileInputText);
        }
        FileDirListView = new ListView(ThisContext);
        FileDirListAdapter = createListAdapter(FileDirList);
        FileDirListView.setAdapter(FileDirListAdapter);
        FileDirListView.setOnItemClickListener(new AdapterView.OnItemClickListener(){
            @Override
            public void onItemClick(AdapterView<?> adapterView, View view, int i, long l) {     //Bei Click auf ListItem
                String olddir = AktDirectory;
                String selitem = "" + FileDirList.get(i);
                if (selitem.charAt(selitem.length()-1) == '/')	selitem = selitem.substring(0, selitem.length()-1);     //Evtl. abschließendes '/' entfernenm wenn ein Verzeichnis gewählt wurde
                //Gesamtpfad zu neuem Verzeichnis oder File erstellen
                if (selitem.equals("..")) {             //Wenn '..' ein Verzeichnis hoch-navigieren
                    AktDirectory = AktDirectory.substring(0, AktDirectory.lastIndexOf("/"));
                } else {                            //Ansonsten in Unterverzeichnis oder gewaehltes File
                    AktDirectory += "/" + selitem;
                }
                SelectedFileDir = DefaultFilename;
                if ((new File(AktDirectory).isFile())) {      //Wenn die Auswahl ein File ist: AktDirectory zurücksetzen auf ursprüngliches Verzeichnis
                    AktDirectory = olddir;
                    SelectedFileDir = selitem;
                }
                UpdateDialogViews();
            }
        });
        InputLayout.addView(FileDirListView);

        //Verzeichnis/File-Auswahl-Dialog erstellen
        AlertDialog.Builder dialogBuilder = new AlertDialog.Builder(ThisContext);
        dialogBuilder.setPositiveButton("OK", new OnClickListener() {       //Wenn OK des FileDialogs gedückt wird
            @Override
            public void onClick(DialogInterface dialog, int which) {
                //Weiterleitung zum registrierten Listener
                if (KEFileDialogListener != null){
                    if (DialogType == FileOpen || DialogType == FileSave) {
                        SelectedFileDir= FileInputText.getText() +"";
                        KEFileDialogListener.onOKClick(AktDirectory + "/" + SelectedFileDir);
                    } else {
                        KEFileDialogListener.onOKClick(AktDirectory);
                    }

                }
            }
        }).setNegativeButton("Cancel", new OnClickListener() {         //Wenn Cancel des FileDialogs gedückt wird
            @Override
            public void onClick(DialogInterface dialog, int which) {
                //Weiterleitung zum registrierten Listener
                KEFileDialogListener.onCancelClick();
            }
        });
        //Layouts zu Dialog zusammenstellen
        dialogBuilder.setView(InputLayout);
        dialogBuilder.setCustomTitle(TitleLayout);
        dialogBuilder.setCancelable(false);
        final AlertDialog FinalDialog = dialogBuilder.create();
        FinalDialog.show();
    }

    //Erzeugt ListenAdapter für ListView
    private ArrayAdapter<String> createListAdapter(List<String> items) {
        return new ArrayAdapter<String>(ThisContext, android.R.layout.select_dialog_item, android.R.id.text1, items) {
            @Override
            public View getView(int position, View convertView, ViewGroup parent) {
                View v = super.getView(position, convertView, parent);
                if (v instanceof TextView) {
                    // Enable list item (directory) text wrapping
                    TextView tv = (TextView) v;
                    tv.getLayoutParams().height = LayoutParams.WRAP_CONTENT;
                    tv.setEllipsize(null);
                }
                return v;
            }
        };
    }

    //Neues Unterverzeichnis erstellen
    private boolean CreateDirectory(String newDir) {
        File newDirFile = new File(newDir);
        if   (! newDirFile.exists() ) return newDirFile.mkdir();
        else return false;
    }

    //Liste mit allen Files und Unterverzeichnissen eines Verzeichnisses erstellen
    private List<String> getFilesAndDirs(String dir) {
        List<String> filedirs = new ArrayList<String>();
        try {
            File dirFile = new File(dir);

            if (! AktDirectory.equals(StartDirectory) ) filedirs.add("..");   //Wenn akt. Verz nicht das root-Vz ist: ".." hinhzufügen als Item um ein Vz hochzunavigieren
            if (! dirFile.exists() || ! dirFile.isDirectory()) return filedirs;     //Wenn File nicht existiert oder kein Verzeichnis ist - leere Rückgabe

            for (File file : dirFile.listFiles()) {     //Alle Files und Directories zur Liste hinzufügen
                if ( file.isDirectory())
                {
                    filedirs.add( file.getName() + "/" );   //Directories schließen mit einem "/"
                } else if (DialogType == FileSave || DialogType == FileOpen) {        //Wenn nur Verzeichnisse angezeiogt werden sollen - Files ignorieren
                    //Ansonsten auch alle Files, die dem Filter-Kriterium entsprechen
                    if (file.getName().substring(file.getName().lastIndexOf(".")).equals(FileExtFilter) || FileExtFilter.equals("*")) filedirs.add( file.getName() );
                }
            }
        } catch (Exception e) {
        }

        //Liste alphabetisch sortieren
        Collections.sort(filedirs, new Comparator<String>() {
            public int compare(String o1, String o2) {
                return o1.compareTo(o2);
            }
        });
        return filedirs;
    }

    //Update der des Dialogtitels, der Liste und der File-Selection
    private void UpdateDialogViews() {
        FileDirList.clear();
        FileDirList.addAll( getFilesAndDirs(AktDirectory) );
        if (DialogType == FileOpen    ) TitleText.setText("Lade " + FileExtFilter + " aus " + AktDirectory);
        if (DialogType == FileSave    ) TitleText.setText("Sichere " + FileExtFilter + " nach " + AktDirectory);
        FileDirListAdapter.notifyDataSetChanged();
        if (DialogType == FileSave || DialogType == FileOpen) {
            FileInputText.setText(SelectedFileDir);
        }
    }


}
