package de.kesystem.remotemessadapter;

import android.support.v4.app.ActivityCompat;
import android.support.v4.content.ContextCompat;
import android.support.v7.app.AppCompatActivity;

import android.Manifest;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothSocket;
import android.content.Intent;
import android.graphics.Color;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Paint;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.view.WindowManager;
import android.widget.Button;
import android.widget.ImageView;
import android.widget.LinearLayout;
import android.widget.Toast;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.UUID;

public class MainActivity extends AppCompatActivity {
    static final String WORKINGDIR = "KE_Messadapter";
    static final String START = "Start";
    static final String STOP = "Stop";
    static final String CONNECT = "Connect";
    static final String DISCONNECT = "Disconnect";
    static final String RECORD = "Record";
    static final String STOPREC = "Stop Record";
    static final byte ConNONE = 0;      //Drei Arten der Connection: Keine BT, File
    static final byte ConFILE = 1;
    static final byte ConBT = 2;
    static final byte Com0 = 0;         //Com-Status
    static final byte Com1 = 1;
    static final byte Com2 = 2;
    static final byte ComSR = -1;
    static final byte STARTBYTE = (byte)0xFF;   //Startbyte in der BT-Übertragung
    static final byte ENDBYTE = (byte)0x00;    //Stopbyte in der BT-Übertragung
    static final long AZUPDATERATE = 250;       //Anzeige-Update-Rate in ms
    static final int HWSRBase = 4;              //Die in der HW eingestellte Basis-Sample-Rate, d.h. Zeit nach der Timer2 überlauft in ms

    //GUI Bedienelemente
    ImageView I_I1, I_I2, I_S1, I_S2, I_V1, I_V2, I_L;
    Button B_Record;
    LinearLayout L_Gesamt, L_Headrow;
    MenuItem M_LoadKonfig, M_Connect, M_StartStop, M_File, M_Bluetooth;

    //Hilfsmittel für GUI Bedienelemente
    Runnable TimerFileStreamTick;                    //Runnable für die Timer-Routine zum Update des Filestreams
    Handler TimerFileStream = new Handler();         //Handler für Timer-Routine zum Update des Filestreams
    Runnable TimerTimeOutTick;                       //Runnable für die Timer-Routine zum Erkennen eines Timeout
    Handler TimerTimeOut = new Handler();            //Handler für Timer-Routine zum Erkennen eines Timeout
    Runnable TimerAzUpdateTick;                      //Runnable für die Timer-Routine zum Update der Anzeige
    Handler TimerAzUpdate = new Handler();           //Handler für Timer-Routine zum Update der Anzeige

    //Bluetooth-Stuff
    static final String keBTNAME = "KE_Messadapt";    //Name des Bluetooth-Moduls, Koppelpasswort=1511
    BluetoothAdapter keBTAdapter;                //Zugriff auf die BT Hardware
    BluetoothDevice keBTDevice;               //Zugriff auf das Device mit entsprechendem Namen
    BluetoothSocket keBTSocket = null;        //Zugriff auf Bluetooth-Connection
    InputStream keBTInputStream;
    OutputStream keBTOutputStream;
    Thread keBTListenThread;
    boolean keBTListening;
    int RXbufsize;                         //Anzahl empfangene Bytes, nach denen onRXReceive ausgelöst wird

    //Spezielle Variablen der Anzeige
    static final Byte RANDL = 1;            //Linker Rand in Anzeigen
    static final Byte RANDR = 1;            //Rechter Rand in Anzeigen
    static final Byte RANDO = 10;            //Oberer Rand in Anzeigen
    static final Byte RANDU = 10;            //Unterer Rand in Anzeigen
    static final Byte LW_KA = 5;           //Linienbreite der Kurven in den Anzeigen
    static final Byte LHGHT = 100;           //Anzeige-Höhe der einzelnen Logiklines

    //Allgemeine Variablen
    String Konfig_Text;                             //Die Daten der geladenen Konfiguration
    String Konfig_File;
    int Konfig_SampleRate;
    int Konfig_BuffSize;
    boolean[] Konfig_KanalAktiv = new boolean[10];
    String[] Konfig_KanalParam = new String[6];
    String[] Konfig_KanalEinheit = new String[6];
    String[] Konfig_KanalText = new String[6];

    List<String> fstream = new ArrayList<>();       //Samplestream fuer Connection mit File
    List<String> pstream = new ArrayList<>();       //Samplestream fuer Aufzeichnung in File
    List<Integer>[] istream = (ArrayList<Integer>[]) new ArrayList[6];      //Rohwertstreams fuer die 6 analogen Kanäle
    List<Double>[] astreama = (ArrayList<Double>[]) new ArrayList[6];      //Anzeigestreams fuer die 6 analogen Kanäle
    List<Byte>[] astreamd = (ArrayList<Byte>[]) new ArrayList[4];           //Anzeigestreams fuer die 4 Logik-Kanäle
    int fspointer;                           //Pointer in Filestream
    int aspointer;                           //Pointer in Anzeigestreams
    byte connected;                         //Connected-Status: Siehe Consts ConNONE, ConFILE, ConBT
    boolean recording;                      //Flag das anzeigt, wenn gerade aufgezeichnet wird
    byte BTComState;                        //Der aktuelle Zustand der Handshake-Kommunikation über BT (z.B. bei Connect, Start, Stop)
    byte BTExpReturn;                        //Als nächstes erwartete Antwort in der Handshake Kommunikation
    boolean stopping;                       //Flag das anzeigt, das nach dem nächsten 10er-Datenpaket der Empfang gestoppt werden soll

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.activity_main);         //Gemaess Manifest wird die App nur im Portrait-Modus betrieben
        //Views aus Layout einbinden
        L_Gesamt = findViewById(R.id.L_Gesamt);
        L_Headrow = findViewById(R.id.L_Headrow);
        I_I1 = findViewById(R.id.Image_I1);
        I_I2 = findViewById(R.id.Image_I2);
        I_S1 = findViewById(R.id.Image_S1);
        I_S2 = findViewById(R.id.Image_S2);
        I_V1 = findViewById(R.id.Image_V1);
        I_V2 = findViewById(R.id.Image_V2);
        I_L = findViewById(R.id.Image_L);
        B_Record = findViewById(R.id.B_Record);
        //Timer für das Streamen von Filestreams
        TimerFileStreamTick = new Runnable() {
            //Timer-Routine, bei Connection mit File-Stream - Update alle 250ms
            @Override
            public void run() { onTimerFileStreamTick(); }
        };
        //Timer für das Erkennen von Timeouts
        TimerTimeOutTick = new Runnable() {
            //Timer-Routine, Zeit bis Ausführung wird in Programm durch setzen des PostDelayed bestimmt
            @Override
            public void run() {
                onTimerTimeOutTick();
            }
        };
        //Timer für das Update der Anzeige
        TimerAzUpdateTick = new Runnable() {
            //Timer-Routine, Zeit bis Ausführung wird in Programm durch setzen des PostDelayed bestimmt
            @Override
            public void run() {
                onTimerAzUpdateTick();
                TimerAzUpdate.postDelayed(this, AZUPDATERATE);
            }
        };

        //Diesen Blödsinn muss man wohl seit SDKxx noch zusätzlich zu dem Manifest machen - User ein mal explizit fragen, ob er Read-Access geben möchte
        if (ContextCompat.checkSelfPermission(this, Manifest.permission.READ_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.READ_EXTERNAL_STORAGE}, 1);
        }
        //Diesen Blödsinn muss man wohl seit SDKxx noch zusätzlich zu dem Manifest machen - User ein mal explizit fragen, ob er Write-Access geben möchte
        if (ContextCompat.checkSelfPermission(this, Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.WRITE_EXTERNAL_STORAGE}, 1);
        }
        //Diesen Blödsinn muss man wohl seit SDKxx noch zusätzlich zu dem Manifest machen - User ein mal explizit fragen, ob er Storage-Access geben möchte
        if (ContextCompat.checkSelfPermission(this, Manifest.permission.MANAGE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
            ActivityCompat.requestPermissions(this, new String[]{Manifest.permission.MANAGE_EXTERNAL_STORAGE}, 1);
        }

        //Hier dürfen keine Referenzen auf die Widgets rein, weil die hier evtl noch nicht erzeigt sind
        //Z.g. Abfrage I_Anzeige.getWidth erzeugt 0 - Menus sind noch nicht mal angelegt, d.h. Referenzen auf Menuitems erzeugen Crash, Erzeugen einer initialen Anzeigegrafik erzeugt Crash...
        //Deswegen wird die eigene Initialisierung der App über eine Runnable verzögert, also nach Anlegen aller Widgets und Menues gestartet
        Runnable App_Init = new Runnable() {
            @Override
            public void run() {
                onStartup();
            }
        };
        Handler App_Init_Handler = new Handler();
        App_Init_Handler.postDelayed(App_Init, 100);       //Eigene Initialisierungen, die evtl schon auf Widgets und Menues zugreifen müssen
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        // Menu erzeugen
        getMenuInflater().inflate(R.menu.main, menu);
        M_LoadKonfig = menu.findItem(R.id.M_LoadKonfig);
        M_Connect = menu.findItem(R.id.M_Connect);
        M_StartStop = menu.findItem(R.id.M_StartStop);
        M_File = menu.findItem(R.id.M_File);
        M_Bluetooth = menu.findItem(R.id.M_Bluetooth);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        //Menupunkte mit Routinen verbinden
        int id = item.getItemId();
        switch (id) {
            case R.id.M_LoadKonfig:
                M_LoadKonfig_Click();
                break;
            case R.id.M_Connect:
                M_Connect_Click();
                break;
            case R.id.M_StartStop:
                M_StartStop_Click();
                break;
            case R.id.M_File:
                M_File_Click();
                break;
            case R.id.M_Bluetooth:
                M_Bluetooth_Click();
                break;
        }
        return super.onOptionsItemSelected(item);
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
        keBTListening = false;                //Bluetooth-Listener-Thread stoppen
        if (keBTAdapter != null) keBTAdapter.disable();     //Bluetooth ausschalten
    }

    public void onStartup() {
        //Startup-Settings
        //In onCreate dürfen keine Referenzen auf die Widgets rein, weil die da evtl noch nicht erzeigt sind
        //Z.g. Abfrage I_Anzeige.getWidth erzeugt 0 - Referenzen auf Menuitems erzeugen Crash, Erzeugen einer initialen Anzeigegrafik erzeugt Crash....
        //Das hier wird durch die Runnable App_Init verzögert gestartet - also nach Erzeugen aller Grundlagen. D.h. das hier sollte auf alles zugreifen dürfen
        M_File.setChecked(true);
        M_Connect.setTitle(CONNECT);
        connected = ConNONE;
        M_StartStop.setTitle(START);
        M_StartStop.setEnabled(false);
        B_Record.setText(RECORD);
        B_Record.setEnabled(true);
        B_Record.setBackgroundColor(Color.RED);
        recording=false;
        TimerFileStream.removeCallbacks(TimerFileStreamTick);         //Disabled Timer
        TimerAzUpdate.removeCallbacks(TimerAzUpdateTick);           //Anzeige-Updates erst nach "Start"
        byte b;
        for (b=0;b<=5;b++) astreama[b] = new ArrayList<>();         //Stream-Arrays initialisieren
        for (b=0;b<=5;b++) istream[b] = new ArrayList<>();
        for (b=0;b<=3;b++) astreamd[b] = new ArrayList<>();
        //Standardkonfiguration: Alle Kanäle aktive, 10 Hz, 10 Sek, 1:1-Anzeige
        Konfig_File="Standardkonfig.kmk";
        Konfig_Text="Standardkonfiguration";
        Konfig_SampleRate=10;
        Konfig_BuffSize=100;
        for (b=0;b<=5;b++) {
            Konfig_KanalAktiv[b]=true;
            Konfig_KanalParam[b]="lin(1:0)";
            Konfig_KanalEinheit[b]=".";
            Konfig_KanalText[b]="1:1-Anzeige";
        }
        for (b=6;b<=9;b++) {
            Konfig_KanalAktiv[b]=true;
        }
        AnzeigenInitialisieren();

        //Arbeitsdirectory erstellen, wenn es noch nicht existiert
        File protdir = new File(Environment.getExternalStorageDirectory(), WORKINGDIR);
        boolean success = true;
        if (!protdir.exists()) success = protdir.mkdirs();
        if (!success)
            Toast.makeText(getApplicationContext(), "Arbeitsverzeichnis konnte nicht erstellt werden", Toast.LENGTH_SHORT).show();
    }

    public void onTimerFileStreamTick() {
        //Wird ausgelöst, wenn der Filestream-Timer auslöst
        onFileReceive();
        //onBTSimReceive();       //Nur für Testzwecke bzgl simuliertem BT-Empfang
        TimerFileStream.postDelayed(TimerFileStreamTick, (long)(Konfig_SampleRate*HWSRBase));  //Nächste Auslösung je nach SampleRate (Konfig_SampleRate*BasisSampleRate)
    }

    public void onTimerTimeOutTick() {
        //Wird ausgelöst, wenn der TimeOut-Timer auslöst
        if (BTComState==Com0) {
            Toast.makeText(MainActivity.this, "TimeOut beim Herstellen der Verbindung (0x00)", Toast.LENGTH_LONG).show();
            Disconnect();
        } else if (BTComState==Com1) {
            Toast.makeText(MainActivity.this, "TimeOut beim Starten (0x01)", Toast.LENGTH_LONG).show();
        } else if (BTComState==Com2) {
            Toast.makeText(MainActivity.this, "TimeOut beim Stoppen (0x02)", Toast.LENGTH_LONG).show();
        }else if (BTComState==ComSR) {
            Toast.makeText(MainActivity.this, "TimeOut beim Herstellen der Verbindung (SR-Wert)", Toast.LENGTH_LONG).show();
            Disconnect();
        } else {
            Toast.makeText(MainActivity.this, "Ungültiger ComState!", Toast.LENGTH_LONG).show();
        }
    }

    public void onTimerAzUpdateTick() {
        //Wird regelmäßig ausgelöst, um Anzeigen neu zu zeichnen
        Display_MakeAll();
    }

    public void M_LoadKonfig_Click() {
        //Menupunkt "Konfiguration laden"

        //FileOpenDialog erzeigen und callback registireren
        KEFileDialog FileOpenDialog = new KEFileDialog(MainActivity.this, "FileOpen",
                new KEFileDialog.KEFileDialogInterface() {
                    @Override
                    public void onOKClick(String chosenDir) {
                        //Wenn OK gedrückt wird
                        BufferedReader reader;
                        byte m;
                        int i;
                        String line;
                        String[] items = new String[4];

                        try {
                            reader = new BufferedReader(new FileReader(chosenDir));
                            line = reader.readLine();       //Header-Zeile
                            if (line.regionMatches(0,"1.0:KMK",0,7)) {
                                Konfig_File=chosenDir.substring(chosenDir.lastIndexOf("/")+1);
                                Konfig_Text=reader.readLine();
                                Konfig_SampleRate=Integer.valueOf(reader.readLine());
                                Konfig_BuffSize=Integer.valueOf(reader.readLine());
                                for (m=0;m<=5;m++) {
                                    line = reader.readLine();   //Jede Zeile beinhaltet 4 Items
                                    items = line.split(";", 4);
                                    Konfig_KanalAktiv[m]=false;
                                    if (Integer.valueOf(items[0])>0) Konfig_KanalAktiv[m]=true;
                                    Konfig_KanalParam[m]=items[1];
                                    Konfig_KanalEinheit[m]=items[2];
                                    Konfig_KanalText[m]=items[3];
                                }
                                line = reader.readLine();   //Aktiv-Status der 4 Logikkanäle, Zeile beinhaltet 4 Items
                                items = line.split(";", 4);
                                for (m=0;m<=3;m++) {
                                    Konfig_KanalAktiv[m+6]=false;
                                    if (Integer.valueOf(items[m])>0) Konfig_KanalAktiv[m+6]=true;
                                }
                                for (m=0;m<6;m++) {                   //6 analoge Anzeige-Streams mit der gewünschten BufferSize generieren
                                    astreama[m].clear();
                                    istream[m].clear();
                                    for (i=0;i<Konfig_BuffSize;i++) {
                                        astreama[m].add(0.0);
                                        istream[m].add(0);
                                    }
                                }
                                for (m=0;m<4;m++) {                   //4 Logik Anzeige-Streams mit der gewünschten BufferSize generieren
                                    astreamd[m].clear();
                                    for (i=0;i<Konfig_BuffSize;i++) {
                                        astreamd[m].add((byte)0);
                                    }
                                }
                                aspointer = 0;
                                M_File.setChecked(false);           //Nach Laden einer Konfiguration, kommt normalerweise die Verbindung über BT - Files bringen ja eigene Konfig mit
                                M_Bluetooth.setChecked(true);
                                AnzeigenInitialisieren();
                            } else {
                                Toast.makeText(MainActivity.this, "Unbekanntes Fileformat", Toast.LENGTH_LONG).show();
                            }
                            reader.close();
                        } catch (IOException e) {
                            e.printStackTrace();
                            Toast.makeText(MainActivity.this, "Fehler beim Lesen" + chosenDir, Toast.LENGTH_LONG).show();
                        }
                    }

                    @Override
                    public void onCancelClick() {
                        //Wenn Cancel gedrückt wird
                    }
                });
        FileOpenDialog.StartDirectory = Environment.getExternalStorageDirectory().getAbsolutePath() + "/" + WORKINGDIR;        //Start-Directory
        FileOpenDialog.DefaultFilename = "";                      //Default-Filename
        FileOpenDialog.FileExtFilter = ".kmk";                        //Extension-Filter
        FileOpenDialog.Generate();
    }

    public void M_File_Click() {
        M_Bluetooth.setChecked(false);
        M_File.setChecked(true);
    }

    public void M_Bluetooth_Click() {
        M_File.setChecked(false);
        M_Bluetooth.setChecked(true);
    }

    public void M_Connect_Click() {
        if (M_Connect.getTitle() == CONNECT) {
            Connect();
            if (connected > ConNONE) {
                SetConnectViewsStatus(true);
            }
        } else {
            Disconnect();
            if (connected == ConNONE) {
                SetConnectViewsStatus(false);
            }
        }
    }

    public void M_StartStop_Click() {
        if (M_StartStop.getTitle() == START) {  //Druck auf Start
            if (connected == ConFILE) {         //Verbindung ist zu File
                TimerFileStream.postDelayed(TimerFileStreamTick, 0);       //Enabled Filestream-Timer
                TimerAzUpdate.postDelayed(TimerAzUpdateTick, AZUPDATERATE);            //Enabled Anzeige-Timer
                M_StartStop.setTitle(STOP);
                M_Connect.setEnabled(false);
            } else {                            //Verbindung ist zu Bluetooth
                keBTListening = true;          //Startet den BT-Listener-Thread
                SendCommand((byte)0x01,Com1,(byte)0x01);    //0x01 ist das Kommando zum Starten der Datenuebertragung, erwartet wird die Rückantwort 0x01 - Rest in onRXReceive oder TimerTimeOutTick
            }
        } else {                                //Druck auf Stop
            if (connected == ConFILE) {         //Verbindung ist zu File
                TimerFileStream.removeCallbacks(TimerFileStreamTick);               //Disabled Fielstream-Timer
                TimerAzUpdate.removeCallbacks(TimerAzUpdateTick);                   //Disabled Anzeige-Timer
                M_StartStop.setTitle(START);
                M_Connect.setEnabled(true);
            } else {                            //Verbindung ist zu Bluetooth
                keBTListening = true;          //Startet den BT-Listener-Thread
                //SendCommand((byte)0x02,Com2,(byte)0x02);    //Cmd-Sendung muss hier von Hand erfolgen, da SendCommand RXbufsize auf 1 setzt, das darf nicht sein, da die Antwort vom PIC ein 10er-Paket (10x 0x02) ist
                stopping=true;                 //Stattdessen wird stopping=true gesetzt - das signalisiert, dass ein Stop-Request gesendet wurde
                BTComState=Com2;
                BTExpReturn=(byte)0x02;
                TimerTimeOut.postDelayed(TimerTimeOutTick, (long)(2000));       //TimeOut nach 2 Sekunde
                keBTwriteByte((byte)0x02);          //0x02 ist das Kommando zum Stoppen der Datenuebertragung, erwartet wird die Rückantwort 10x 0x02 - Rest in onRXReceive oder TimerTimeOutTick
            }
        }
    }

    public void B_Record_Click(View view) {
        //Druck auf Record-Button (gelinkt direkt im MainActivity.xml)
        if (B_Record.getText() == RECORD) {
            B_Record.setText(STOPREC);
            recording=true;
            B_Record.setBackgroundColor(Color.GREEN);
            //Protokollstream neu aufsetzen, aktuellen Anzeigebuffer als History direkt in den Protokollstream schreiben
            pstream.clear();
            int i;
            byte b;
            String line;
            i=aspointer;    //aspointer zeigt auf ältesten Wert im Anzeigebuffer
            while (i < Konfig_BuffSize) {   //Zuerst Buffer von aspointer bis Ende...
                line = "";
                for (b=0;b<6;b++) {         //Analogkanäle
                    if (Konfig_KanalAktiv[b]) line = line + istream[b].get(i) + ";";
                }
                for (b=0;b<4;b++) {         //Logikkanäle
                    if (Konfig_KanalAktiv[b+6]) line = line + astreamd[b].get(i) + ";";
                }
                line=line.substring(0,line.length()-1);     //Letzten ";" eliminieren
                pstream.add(line);
                i++;
            }
            i=0;
            while (i < aspointer) {         //...dann von 0 bis neuestem Wert - nicht einfach Konfig_Buffsize Werte übernehmen, evtl zählt aspointer während der Aktion hoch
                line = "";
                for (b=0;b<6;b++) {         //Analogkanäle
                    if (Konfig_KanalAktiv[b]) line = line + istream[b].get(i) + ";";
                }
                for (b=0;b<4;b++) {         //Logikkanäle
                    if (Konfig_KanalAktiv[b+6]) line = line + astreamd[b].get(i) + ";";
                }
                line=line.substring(0,line.length()-1);     //Letzten ";" eliminieren
                pstream.add(line);
                i++;
            }
        } else {
            B_Record.setText(RECORD);
            ProtStreamFileSchreiben();          //Protokollstream in File abspeichern - Die Prüfung, ob es überhaupt einen Stream gibt, findet in der Schreib-Routine statt - z.B. wenn man nur Cheched und wieder Unchecked, ohne einen Stream per Connect ablaufen zu lassen
            recording=false;
            B_Record.setBackgroundColor(Color.RED);
        }
    }

    private void onFileReceive() {
        //Von TimerFileStreamTick-Routine ausgelöst - bei Connection mit File-Stream - Update alle 250ms
        //Entnimmt ein Sample aus dem Filestream und fügt diesen in den Anzeigestream ein
        int n, IntWert;
        byte b;
        String samplestring;
        String[] stritems = new String[10];      //Pro Zeile kann es bis zu 10 Werte geben
        double Anzeigewert;

        samplestring = fstream.get(fspointer);     //Einen Sample-String aus dem Filestream laden (eine Zeile der Filestream-Liste mit bis zu 10 Werten)
        if (recording) pstream.add(samplestring);   //Wenn aufgezeichnet werden soll: Samplestring in Protokoll-stream kopieren
        fspointer++;
        if (fspointer == fstream.size()) fspointer = 0;
        //Frame-String umwandeln in Int-Sequenz
        samplestring = samplestring.replace(" ", "");    //Falls irgendwo Leerzeichen drin sind - Integer.valueOf kommt damit nicht klar
        stritems = samplestring.split(";", 10);
        b=0;
        for (n=0;n<=5;n++) {
            if (Konfig_KanalAktiv[n]) {             //Jedem aktiven Kanal einen Sample-Wert zuweisen, inaktive Kanäle bleiben auf 0
                IntWert=Integer.valueOf(stritems[b]);
                istream[n].set(aspointer,IntWert);
                Anzeigewert=getAnzeigewert(IntWert,Konfig_KanalParam[n]);
                astreama[n].set(aspointer,Anzeigewert);
                b++;
            }
        }
        for (n=6;n<=9;n++) {
            if (Konfig_KanalAktiv[n]) {             //Jedem aktiven Kanal einen Sample-Wert zuweisen, inaktive Kanäle bleiben auf 0
                astreamd[n-6].set(aspointer,Byte.valueOf(stritems[b]));
                b++;
            }
        }
        aspointer++;
        if (aspointer == astreama[0].size()) aspointer = 0;
    }

    private void onBTSimReceive() {
        //Wird vom File-Stream-Timer ausgelöst - zum Test von onRXReceive ohne wirklich eine BT-Verbindung zu haben
        //D.h. generiert ein 10-Byte Rxpacket und schleust das nach onRXReceive ein
        byte[] TestVektor = new byte[10];

        //Testvektor - ergibt nach Kanalzerlegung die Werte 11,74,992,993,35,173 und Logik 0,1,1,0
        TestVektor[0]=STARTBYTE;
        TestVektor[1]=11;
        TestVektor[2]=40;
        TestVektor[3]=1;
        TestVektor[4]=126;
        TestVektor[5]=-8;
        TestVektor[6]=35;
        TestVektor[7]=-76;
        TestVektor[8]=98;
        TestVektor[9]=ENDBYTE;
        RXbufsize=10;
        onRXReceive(TestVektor);
    }

    public void onRXReceive(byte[] RXpacket) {
        // Das hier wird vom keBTListen-Thread ausgelöst, wenn über RXbufsize eingestellte Anzahl von Bytes über Bluetooth empfangen wurden
        // Es gibt zwei Modi: 1. Kommando-Austausch in dem einzelne Bytes übertragen werden, 2. Run-Modus in dem immer 10 Bytes im RXpacket liegen

        if (RXpacket.length==1) {               //Modus 1, einzelnes Byte mit Kommandowert
            TimerTimeOut.removeCallbacks(TimerTimeOutTick);      //Disabled TimeOut
            ExecuteReturn(RXpacket[0]);                         //Empfangene Antwort verwerten
        } else if (RXpacket.length==10) {       //Modus 2, 10er-Pakete mit Anzeigedaten
            int n;
            byte b;
            double Anzeigewert;
            int[] intitems = new int[10];

            //Empfangene Byte-Werte in Int umrechnen, Jave-Byte sind immer signed - d.h. Umrechnung von (-128 bis +127) auf (0 bis 255) ist nötwenig (0x00=0 / 0x7F=127 / 0x80=-128 / 0xFF=-1)
            for (n=1;n<9;n++) {             //Nur für Werte-Bytes, Start- und Stopbyte müssen nicht umgerechnet werden
                if (RXpacket[n]>=0) {
                    intitems[n] = RXpacket[n];
                } else {
                    intitems[n] = 256 + RXpacket[n];
                }
            }
            if (RXpacket[0]==STARTBYTE & RXpacket[9]==ENDBYTE) {        //Startbyte und Stopbyte prüfen
                //Einzelne Bytes in ihre Kanalwerte zerlegen
                n = (intitems[2] & 0x03) * 256 + intitems[1];           //I1 = Byte2(1..0),Byte1(7..0)
                istream[0].set(aspointer,n);
                n = (intitems[3] & 0x0F) * 64 + (intitems[2] >> 2);    //I2 = Byte3(3..0),Byte2(7..2)
                istream[1].set(aspointer,n);
                n = (intitems[4] & 0x3F) * 16 + (intitems[3] >> 4);    //S1 = Byte4(5..0),Byte3(7..4)
                istream[2].set(aspointer,n);
                n = intitems[5] * 4 + (intitems[4] >> 6);             //S2 = Byte5(7..0),Byte4(7..6)
                istream[3].set(aspointer,n);
                n = (intitems[7] & 0x03) * 256 + intitems[6];           //V1 = Byte7(1..0),Byte6(7..0)
                istream[4].set(aspointer,n);
                n = (intitems[8] & 0x0F) * 64 + (intitems[7] >> 2);    //V2 = Byte8(3..0),Byte7(7..2)
                istream[5].set(aspointer,n);
                if ((intitems[8] & 0x10) == 0x10) astreamd[0].set(aspointer,(byte)1); else astreamd[0].set(aspointer,(byte)0);   //L4, L3, L2, L1 = Byte8(7..4)
                if ((intitems[8] & 0x20) == 0x20) astreamd[1].set(aspointer,(byte)1); else astreamd[1].set(aspointer,(byte)0);
                if ((intitems[8] & 0x40) == 0x40) astreamd[2].set(aspointer,(byte)1); else astreamd[2].set(aspointer,(byte)0);
                if ((intitems[8] & 0x80) == 0x80) astreamd[3].set(aspointer,(byte)1); else astreamd[3].set(aspointer,(byte)0);
                //Wenn aufgezeichnet werden soll: Samplestring erzeugen und in Protokoll-stream kopieren
                if (recording) {
                    String line = "";
                    for (b=0;b<6;b++) {         //Analogkanäle
                        if (Konfig_KanalAktiv[b]) line = line + istream[b].get(aspointer) + ";";
                    }
                    for (b=0;b<4;b++) {         //Logikkanäle
                        if (Konfig_KanalAktiv[b+6]) line = line + astreamd[b].get(aspointer) + ";";
                    }
                    line=line.substring(0,line.length()-1);     //Letzten ";" eliminieren
                    pstream.add(line);
                }
                //Analoge Anzeigestreams um den jeweiligen Anzeigewert erweitern, Logik-Anzeigestreams wurden schon gesetzt
                for (b=0;b<6;b++) {
                    if (Konfig_KanalAktiv[b]) {             //Jedem aktiven Kanal einen Sample-Wert zuweisen, inaktive Kanäle bleiben auf 0
                        Anzeigewert=getAnzeigewert(istream[b].get(aspointer),Konfig_KanalParam[b]);
                        astreama[b].set(aspointer,Anzeigewert);
                    }
                }
                aspointer++;
                if (aspointer == astreama[0].size()) aspointer = 0;
            } else if (RXpacket[0]==0x02 & RXpacket[9]==0x02) {         //Ein Antwort-Frame auf einen Stop-request
                if (stopping==true) {
                    TimerTimeOut.removeCallbacks(TimerTimeOutTick);      //Disabled TimeOut
                    ExecuteReturn((byte)0x02);      //Com-Status sollte Com2 sein
                } else {
                    Toast.makeText(MainActivity.this, "Stop-response-Frame erhalten ohne dass stopping=true...", Toast.LENGTH_LONG).show();
                }
            } else {
                Toast.makeText(MainActivity.this, "Sync-Fehler!!!", Toast.LENGTH_LONG).show();
            }
        } else {        //Unbekannter Modus, weder 1 noch 10 Bytes
            Toast.makeText(MainActivity.this, "Unbekannter Com-Modus!!!", Toast.LENGTH_LONG).show();
        }
    }

    public void SetConnectViewsStatus(boolean connected) {
        //Stellt den enabled-Status der Views ein - eigentlich ein Thema für B_Connect_Click, muss aber auch separat in Fileconnect ausgeführt werden
        if (connected) {
            M_Connect.setTitle(DISCONNECT);
            M_StartStop.setEnabled(true);
            M_LoadKonfig.setEnabled(false);
            M_File.setEnabled(false);
            M_Bluetooth.setEnabled(false);
            getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);	//Waehrend eines Streams soll der Bildschirm an bleiben
        } else {
            M_Connect.setTitle(CONNECT);
            M_StartStop.setEnabled(false);
            M_LoadKonfig.setEnabled(true);
            M_File.setEnabled(true);
            M_Bluetooth.setEnabled(true);
            getWindow().clearFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);	//Ausserhalb eines Streams kann der Bildschirm ruhig automatisch ausgehen
        }
    }

    public void ExecuteReturn (byte retwert) {
        //Abarbeitung einer BT-Antwort entsprechend ComState

        if (BTComState==Com0) {             //Kommando zur Übertragung des SR-Werts
            if (retwert == BTExpReturn) {       //Gesendet wurde 0x00, Antwort=0x00: Alles Korrekt
                SendCommand((byte)Konfig_SampleRate, ComSR, (byte)Konfig_SampleRate);   //SR-Wert senden
            } else {
                Toast.makeText(MainActivity.this, "Falschantwort bei Verbindung (Com0): "+ retwert +" statt " + BTExpReturn, Toast.LENGTH_LONG).show();
                Disconnect();
            }
        } else if (BTComState==Com1) {
            if (retwert == BTExpReturn) {       //Gesendet wurde 0x01, Antwort=0x01: Alles Korrekt
                RXbufsize=10;               //Ab sofort 10er-Pakete empfangen
                TimerAzUpdate.postDelayed(TimerAzUpdateTick, AZUPDATERATE);            //Enabled Anzeige-Timer
                M_StartStop.setTitle(STOP);
                M_Connect.setEnabled(false);
                stopping=false;
                keBTwriteByte((byte)0x01);          //Alles ist aufgesetzt, finales 0x01 senden - darauf gibt es vom PIC keine Antwort, sondern er legt direkt mit der Übertragung der 10er-Pakete los
            } else {
                Toast.makeText(MainActivity.this, "Falschantwort bei Start (Com1): "+ retwert +" statt " + BTExpReturn, Toast.LENGTH_LONG).show();
            }
        } else if (BTComState==Com2) {
            if (retwert == BTExpReturn) {       //Gesendet wurde 0x02, Antwort=0x02: Alles Korrekt
                TimerAzUpdate.removeCallbacks(TimerAzUpdateTick);                   //Disabled Anzeige-Timer
                Toast.makeText(MainActivity.this, "Ablauf gestoppt", Toast.LENGTH_LONG).show();
                M_StartStop.setTitle(START);
                M_Connect.setEnabled(true);
                RXbufsize=1;                //Ab sofort einzelne Bytes empfangen
            } else {
                Toast.makeText(MainActivity.this, "Falschantwort bei Stop (Com2): "+ retwert +" statt " + BTExpReturn, Toast.LENGTH_LONG).show();
            }
        }else if (BTComState==ComSR) {
            if (retwert == BTExpReturn) {       //Gesendet wurde SR-Wert, Antwort=SR-Wert: Alles Korrekt
                Toast.makeText(MainActivity.this, "Verbindung hergestellt", Toast.LENGTH_LONG).show();
                connected=ConBT;
                SetConnectViewsStatus(true);
            } else {
                Toast.makeText(MainActivity.this, "Falschantwort bei Verbindung (ComSR): "+ retwert +" statt " + BTExpReturn, Toast.LENGTH_LONG).show();
                Disconnect();
            }
        } else {
            Toast.makeText(MainActivity.this, "Ungültiger ComState!", Toast.LENGTH_LONG).show();
        }
    }

    private void Connect() {
        //Verbindet mit File oder Bluetooth

        if (M_File.isChecked()) {          //Connect mit File
            //FileOpenDialog erzeiugen und callbacl regisitrieren
            KEFileDialog FileOpenDialog = new KEFileDialog(MainActivity.this, "FileOpen",
                    new KEFileDialog.KEFileDialogInterface() {
                        @Override
                        public void onOKClick(String chosenDir) {
                            //Wenn OK gedrückt wird
                            BufferedReader reader;
                            String line;
                            String[] items = new String[10];
                            byte m;
                            int i;

                            try {           //Stream-Files nicht auf binary Data umstellen!!! Java ist extrem umständlich mit binary data, string-encodierung, etc. Bytes werden nur bis 127 korrekt gelesen, darüber kommt nut Müll, etc...
                                reader = new BufferedReader(new FileReader(chosenDir));
                                line = reader.readLine();       //Header-Zeile
                                if (line.regionMatches(0,"1.0:KMM",0,7)) {
                                    line = reader.readLine();       //Erläuterungstext und Dateiname der Konfiguration - 2 Text-Objekte
                                    items = line.split(";", 2);
                                    Konfig_Text=items[0];
                                    Konfig_File=items[1];
                                    line = reader.readLine();       //SampleRate-Byte
                                    Konfig_SampleRate = Integer.parseInt(line);
                                    line = reader.readLine();       //Buffersize-Byte
                                    Konfig_BuffSize = Integer.parseInt(line);
                                    for (m=0;m<=5;m++) {                            //6 Kanal-Konfigurations-Zeilen
                                        line = reader.readLine();                   //4 text-Objekte pro Zeile
                                        items = line.split(";", 4);
                                        Konfig_KanalAktiv[m]=false;
                                        if (Integer.parseInt(items[0])>0) Konfig_KanalAktiv[m]=true;
                                        Konfig_KanalParam[m]=items[1];
                                        Konfig_KanalEinheit[m]=items[2];
                                        Konfig_KanalText[m]=items[3];
                                    }
                                    line = reader.readLine();       //Konfigurationen der Logik-Kanäle - 4 Text-Objekte
                                    items = line.split(";", 4);
                                    for (m=0;m<=3;m++) {
                                        Konfig_KanalAktiv[m+6]=false;
                                        if (Integer.parseInt(items[m])>0) Konfig_KanalAktiv[m+6]=true;
                                    }
                                    fstream.clear();
                                    line = reader.readLine();                       //Jede Zeile ist ein Sample
                                    while (line != null) {                          //Samples bis zum Ende lesen
                                        fstream.add(line);
                                        line = reader.readLine();
                                    }
                                    fspointer = 0;
                                    for (m=0;m<6;m++) {                   //6 analoge Anzeige-Streams und 6 Rohwertstreams mit der gewünschten BufferSize generieren
                                        astreama[m].clear();
                                        istream[m].clear();
                                        for (i=0;i<Konfig_BuffSize;i++) {
                                            astreama[m].add(0.0);
                                            istream[m].add(0);
                                        }
                                    }
                                    for (m=0;m<4;m++) {                   //4 Logik Anzeige-Streams mit der gewünschten BufferSize generieren
                                        astreamd[m].clear();
                                        for (i=0;i<Konfig_BuffSize;i++) {
                                            astreamd[m].add((byte)0);
                                        }
                                    }
                                    aspointer = 0;
                                    connected = ConFILE;                    //Leider können Dialog in Android nicht modal ausgeführt werden, d.h. das hier passiert erst, wenn der
                                    SetConnectViewsStatus(true);   //Nutzer wirklich ein File gewählt hat. Die Aktivitäten, die ansonszten in der B_Connect_Click
                                    //Funktion stehen (und von connected abhängig sind) wurden schon vorher (und damit mit connected=0)
                                    //abgearbeitet... - Deswegen hier nochmal explizit
                                    AnzeigenInitialisieren();
                                } else {
                                    Toast.makeText(MainActivity.this, "Unbekanntes Fileformat", Toast.LENGTH_LONG).show();
                                }
                                reader.close();
                            } catch (IOException e) {
                                e.printStackTrace();
                                Toast.makeText(MainActivity.this, "Fehler beim Lesen" + chosenDir, Toast.LENGTH_LONG).show();
                            }
                        }

                        @Override
                        public void onCancelClick() {
                            //Wenn Cancel gedrückt wird
                        }
                    });
            FileOpenDialog.StartDirectory = Environment.getExternalStorageDirectory().getAbsolutePath() + "/" + WORKINGDIR;        //Start-Directory
            FileOpenDialog.DefaultFilename = "";                      //Default-Filename
            FileOpenDialog.FileExtFilter = ".kmm";                        //Extension-Filter
            FileOpenDialog.Generate();
        } else {                                            //Connect mit Bluetooth
            keBTAdapter = keBTGetAdapter();
            keBTDevice = keBTGetDevice(keBTAdapter, keBTNAME);
            keBTSocket = keBTConnectToSocket(keBTDevice);
            if (keBTSocket != null) {
                keBTOutputStream = keBTGetOutputStream(keBTSocket);
                keBTInputStream = keBTGetInputStream(keBTSocket);
                keBTStartListening();           //Startet den BT-Listener-Thread
                SendCommand((byte)0x00,Com0,(byte)0x00);    //0x00 ist das Kommando zum Setzen der SampleRate, erwartet wird die Rückantwort 0x00 - Rest passiert dann in onRXReceice oder in TimerTimeOutTick
            }
        }
    }

    private void SendCommand(byte cmdbyte, byte comstat, byte expret) {
        //Sendet ein Kommando-Byte an das Gerät, inkl. setzen des TimeOuts, der erwarteten Returnantwort
        //comstat ist aktueller Com-State, expret ist erwartete Antwort
        RXbufsize=1;                    //Es wird jeweils ein Byte als Antwort erwartet
        BTComState=comstat;
        BTExpReturn=expret;
        TimerTimeOut.postDelayed(TimerTimeOutTick, (long)(1000));       //TimeOut nach 1 Sekunde
        keBTwriteByte(cmdbyte);          //Kommando senden - Rest passiert dann in onRXReceice oder in TimerTimeOutTick
    }

    private void Disconnect() {
        if (M_File.isChecked()) {          //Connect mit File
            connected = ConNONE;
        } else {                                            //Connect mit Bluetooth
            try {
                keBTSocket.close();
                keBTListening = false;      //Stoppt den BT-Listener-Thread
                //keBTAdapter.disable();        //Bluetooth noch eingeschaltet lassen - wird dann in onDestroy ausgeschaltet
                connected = ConNONE;
            } catch (IOException e2) {
                Toast.makeText(getApplicationContext(), "Fehler bei der Trennung", Toast.LENGTH_SHORT).show();
            }
        }
    }

    private void Display_MakeAll() {
        //Generiert die Anzeigen aller aktiven Messkanäle
        double max, min;
        byte n, anzlkan;

        for (n=0;n<=5;n++) {        //Die 6 analogen Anzeigen
            if (Konfig_KanalAktiv[n]) {
                min=GetMinInKanal(n);
                max=GetMaxInKanal(n);
                if (min == max) {
                    min=min-1;
                    max=max+1;
                }
                AnalogDisplay_Generate(n, min, max);
            }
        }
        anzlkan=0;
        for (n=6;n<=9;n++) {
            if (Konfig_KanalAktiv[n]) anzlkan++;
        }
        if (anzlkan>0) LogikDisplay_Generate(anzlkan);
    }

    private void AnalogDisplay_Generate(byte kan, double Ymin, double Ymax) {
        //Erzeugt eine Grafik mit Umrandung und Messwertkurve
        int n, asp;
        double YRange;
        Canvas canvas;
        Bitmap bi;
        int biHeight, biWidth;
        double PixProYVal;                      //y-Skalierung Pixel/Value
        int Ynull;                              //Die y-Koordinate der Anzeige, bei der die Null-Linie liegt (liegt evtl außerehalb der Anzeige, wenn z.B. nur ein positiver oder nur ein negativer Bereich angezeigt werden soll)
        int backcolor;
        Paint scalepaint = new Paint();

        if (kan==0) {
            biHeight = I_I1.getHeight() - 0;           //Volle Höhe ausnutzen
            biWidth = I_I1.getWidth() - 0;             //Volle Breite ausnuten
        } else if (kan==1) {
            biHeight = I_I2.getHeight() - 0;
            biWidth = I_I2.getWidth() - 0;
        } else if (kan==2) {
            biHeight = I_S1.getHeight() - 0;
            biWidth = I_S1.getWidth() - 0;
        } else if (kan==3) {
            biHeight = I_S2.getHeight() - 0;
            biWidth = I_S2.getWidth() - 0;
        } else if (kan==4) {
            biHeight = I_V1.getHeight() - 0;
            biWidth = I_V1.getWidth() - 0;
        } else if (kan==5) {
            biHeight = I_V2.getHeight() - 0;
            biWidth = I_V2.getWidth() - 0;
        } else {
            biHeight = 1;
            biWidth = 1;
        }

        bi = Bitmap.createBitmap(biWidth, biHeight, Bitmap.Config.ARGB_8888);
        canvas = new Canvas(bi);

        YRange = Ymax - Ymin;
        PixProYVal = (biHeight - RANDO - RANDU) / YRange;           //Y-Skalierung Pixel/Value
        Ynull = (int) Math.round(PixProYVal * Ymax) + RANDO;        //Ymax liegt an oberem Anzeigerand, 0-Linie entsprechend ober- oder unterhalb

        backcolor = Color.WHITE;
        scalepaint.setColor(0xFF000000);    //Colorcode für Schwarz
        scalepaint.setStrokeWidth(3);
        scalepaint.setTextSize(80);
        //Bitmap füllen
        canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
        canvas.drawLine(1, 1, biWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
        canvas.drawLine(biWidth-1, 1, biWidth-1, biHeight - 1, scalepaint);
        canvas.drawLine(biWidth-1, biHeight - 1, 1, biHeight - 1, scalepaint);
        canvas.drawLine(1, biHeight - 1, 1, 1, scalepaint);

        //Fügt die Werte von astream als Messkurve in den Graphen ein, aspointer zwigt jeweils auf dem ältesten Wert, aspointer-1 ist der jüngste
        Paint penline = new Paint();
        int oldx, oldy, newx, newy;

        penline.setColor(Color.BLUE);
        penline.setStrokeWidth(LW_KA);
        //Kurve in aktuelles Bitmap einzeichnen
        oldx = RANDL;
        oldy = Ynull - (int) (astreama[kan].get(aspointer) * PixProYVal);
        asp=aspointer+1;
        if (asp == astreama[kan].size()) asp = 0;
        for (n = 1; n < astreama[kan].size(); n++) {
            newx = (int) (Math.round((double) n / (double) astreama[kan].size() * (double) (biWidth - RANDL - RANDR))) + RANDL;
            newy = Ynull - (int) (astreama[kan].get(asp) * PixProYVal);
            canvas.drawLine(oldx, oldy, newx, newy, penline);
            oldx = newx;
            oldy = newy;
            asp++;
            if (asp == astreama[kan].size()) asp = 0;
        }

        asp=aspointer-1;        //asp auf neuesten Wert setzen
        if (asp<0) asp=astreama[kan].size()-1;
        if (kan==0) {
            canvas.drawText("I1 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_I1.setImageBitmap(bi);
        } else if (kan==1) {
            canvas.drawText("I2 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_I2.setImageBitmap(bi);
        } else if (kan==2) {
            canvas.drawText("S1 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_S1.setImageBitmap(bi);
        } else if (kan==3) {
            canvas.drawText("S2 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_S2.setImageBitmap(bi);
        } else if (kan==4) {
            canvas.drawText("V1 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_V1.setImageBitmap(bi);
        } else if (kan==5) {
            canvas.drawText("V2 = " + String.format("%.2f", astreama[kan].get(asp)) + Konfig_KanalEinheit[kan],100,100,scalepaint);
            I_V2.setImageBitmap(bi);
        }
    }

    private void LogikDisplay_Generate(byte anzlkan) {
        //Erzeugt die Logik-Grafik mit Umrandung und bis zu 4 Logikverlaeufen
        int n, asp;
        byte kan, aktkan;
        double YRange;
        Canvas canvas;
        Bitmap bi;
        int biHeight, biWidth, kanHeight;
        double PixProYVal;                      //y-Skalierung Pixel/Value
        int Ylow;                              //Die y-Koordinate in der Anzeige, bei der der Low-Level der aktuellen Logikkurve liegt
        int backcolor;
        Paint scalepaint = new Paint();



        biHeight = I_L.getHeight() - 0;           //Volle Höhe ausnutzen
        biWidth = I_L.getWidth() - 0;             //Volle Breite ausnuten
        kanHeight = (int) Math.round((biHeight-RANDU) / anzlkan) - RANDO; //Höhe der einzelnen Logik-Kurven

        bi = Bitmap.createBitmap(biWidth, biHeight, Bitmap.Config.ARGB_8888);
        canvas = new Canvas(bi);

        backcolor = Color.WHITE;
        scalepaint.setColor(0xFF000000);    //Colorcode für Schwarz
        scalepaint.setStrokeWidth(3);
        //Bitmap füllen
        canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
        canvas.drawLine(1, 1, biWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
        canvas.drawLine(biWidth-1, 1, biWidth-1, biHeight - 1, scalepaint);
        canvas.drawLine(biWidth-1, biHeight - 1, 1, biHeight - 1, scalepaint);
        canvas.drawLine(1, biHeight - 1, 1, 1, scalepaint);

        //Fügt die Werte von astreamd als Logikverlaeufe in den Graphen ein, aspointer zeigt jeweils auf dem ältesten Wert, aspointer-1 ist der jüngste
        Paint penline = new Paint();
        int oldx, oldy, newx, newy;

        penline.setColor(Color.BLUE);
        penline.setStrokeWidth(LW_KA);
        //Kurven in aktuelles Bitmap einzeichnen
        aktkan=0;
        for (kan=0;kan<=3;kan++) {
            if (Konfig_KanalAktiv[kan+6]) {
                aktkan++;
                Ylow = aktkan * (kanHeight+RANDO);
                oldx = RANDL;
                oldy = Ylow - (astreamd[kan].get(aspointer) * kanHeight);
                asp=aspointer+1;
                if (asp == astreama[kan].size()) asp = 0;
                for (n = 1; n < astreamd[kan].size(); n++) {
                    newx = (int) (Math.round((double) n / (double) astreamd[kan].size() * (double) (biWidth - RANDL - RANDR))) + RANDL;
                    newy = Ylow - (astreamd[kan].get(asp) * kanHeight);
                    canvas.drawLine(oldx, oldy, newx, newy, penline);
                    oldx = newx;
                    oldy = newy;
                    asp++;
                    if (asp == astreamd[kan].size()) asp = 0;
                }
            }
        }
        I_L.setImageBitmap(bi);
    }

    private double getAnzeigewert(int Roh, String KanPar) {
        //rechnet den Integer-Rohwert um in einen Double-Anzeigewert, gemaess der Umrechnugnsanweisung in KanPar
        String funkt, params;
        String[] items = new String[2];
        double erg;

        //KanPar sieht z.B. so aus: lin(3,987;-2,345)
        funkt = KanPar.substring(0,3).toLowerCase();        //Die gewünschte Funktion steht in den ersten 3 Zeichen
        params = KanPar.substring(4,KanPar.length()-1);     //Die Parameter stehen innerhalb der Klammern
        params = params.replace(',','.');     //parseDouble kommt nicht mit "," klar, wie es z.B. der PC erstellt....
        params = params.replace(" ", "");    //Falls irgendwo Leerzeichen drin sind - parseDouble kommt damit nicht klar
        if (funkt.equals("lin")) {
            items = params.split(":", 2);       //Lineare Berechnungen haben zwei Parameter
            double stg, off;
            stg = Double.parseDouble(items[0]);
            off = Double.parseDouble(items[1]);
            erg = stg * Roh + off;
        } else if (funkt.equals("log")) {
            items = params.split(":", 2);       //Logarithmus Berechnungen haben zwei Parameter
            double mlt, off;
            mlt = Double.parseDouble(items[0]);
            off = Double.parseDouble(items[1]);
            erg = mlt * Math.log(Roh) + off;
        } else if (funkt.equals("exp")) {
            items = params.split(":", 2);       //Exponent-Berechnungen haben zwei Parameter
            double mlt, ex;
            mlt = Double.parseDouble(items[0]);
            ex = Double.parseDouble(items[1]);
            erg = mlt * Math.exp(ex*Roh);
        } else if (funkt.equals("pot")) {
            items = params.split(":", 2);       //Potenz Berechnungen haben zwei Parameter
            double mlt, pot;
            mlt = Double.parseDouble(items[0]);
            pot = Double.parseDouble(items[1]);
            erg = mlt * Math.pow(Roh,pot);
        } else {
            erg = 0.0;
        }
        return erg;
    }

    private double GetMinInKanal(byte kan) {
        //Ermittelt den Minimum-wert eines analogen kanals
        int m;
        double x;

        x = astreama[kan].get(0);
        for (m = 1; m < astreama[kan].size(); m++) {
            if (astreama[kan].get(m) < x) x = astreama[kan].get(m);
        }
        return (x);
    }

    private double GetMaxInKanal(byte kan) {
        //Ermittelt den Maximum-wert eines analogen kanals
        int m;
        double x;

        x = astreama[kan].get(0);
        for (m = 1; m < astreama[kan].size(); m++) {
            if (astreama[kan].get(m) > x) x = astreama[kan].get(m);
        }
        return (x);
    }

    private void AnzeigenInitialisieren() {
        //Initialisiert die einzelnen Anzeigen, blendet je nach KanalAktivität einzelne Anzeigen ein oder aus und stellt deren Höhe ein
        byte anzdkan, anzakan, n;
        int dHeight, aHeight, xWidth;
        Canvas canvas;
        Bitmap bi;
        int backcolor;
        Paint scalepaint = new Paint();

        anzakan=0;
        for (n=0;n<=5;n++) {
            if (Konfig_KanalAktiv[n]) anzakan++;
        }
        anzdkan=0;
        for (n=6;n<=9;n++) {
            if (Konfig_KanalAktiv[n]) anzdkan++;
        }

        dHeight=anzdkan*(LHGHT+RANDU) + RANDO;     //Die einzelnen Logik-Verläufe sollen eine Höhe von 20 Pixeln haben
        aHeight=(L_Gesamt.getHeight()-L_Headrow.getHeight()-dHeight) / anzakan;     //Die analogen Anzeigen sollen die restliche Hoehe im Layout einnehmen

        ViewGroup.LayoutParams aLP;
        backcolor = Color.WHITE;
        scalepaint.setColor(0xFF000000);    //Colorcode für Schwarz
        scalepaint.setStrokeWidth(3);
        scalepaint.setTextSize(80);

        aLP=I_I1.getLayoutParams();
        if (Konfig_KanalAktiv[0]) {
            I_I1.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_I1.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("I1",100,100,scalepaint);
            I_I1.setImageBitmap(bi);
        } else {
            I_I1.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_I1.setLayoutParams(aLP);

        aLP=I_I2.getLayoutParams();
        if (Konfig_KanalAktiv[1]) {
            I_I2.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_I2.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("I2",100,100,scalepaint);
            I_I2.setImageBitmap(bi);
        } else {
            I_I2.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_I2.setLayoutParams(aLP);

        aLP=I_S1.getLayoutParams();
        if (Konfig_KanalAktiv[2]) {
            I_S1.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_S1.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("S1",100,100,scalepaint);
            I_S1.setImageBitmap(bi);
        } else {
            I_S1.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_S1.setLayoutParams(aLP);

        aLP=I_S2.getLayoutParams();
        if (Konfig_KanalAktiv[3]) {
            I_S2.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_S2.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("S2",100,100,scalepaint);
            I_S2.setImageBitmap(bi);
        } else {
            I_S2.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_S2.setLayoutParams(aLP);

        aLP=I_V1.getLayoutParams();
        if (Konfig_KanalAktiv[4]) {
            I_V1.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_V1.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("V1",100,100,scalepaint);
            I_V1.setImageBitmap(bi);
        } else {
            I_V1.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_V1.setLayoutParams(aLP);

        aLP=I_V2.getLayoutParams();
        if (Konfig_KanalAktiv[5]) {
            I_V2.setVisibility(View.VISIBLE);
            aLP.height=aHeight;
            xWidth=I_V2.getWidth();
            bi = Bitmap.createBitmap(xWidth, aHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, aHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, aHeight - 1, 1, aHeight - 1, scalepaint);
            canvas.drawLine(1, aHeight - 1, 1, 1, scalepaint);
            canvas.drawText("V2",100,100,scalepaint);
            I_V2.setImageBitmap(bi);
        } else {
            I_V2.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_V2.setLayoutParams(aLP);

        aLP=I_L.getLayoutParams();
        if (anzdkan > 0) {
            I_L.setVisibility(View.VISIBLE);
            aLP.height=dHeight;
            xWidth=I_L.getWidth();
            bi = Bitmap.createBitmap(xWidth, dHeight, Bitmap.Config.ARGB_8888);
            canvas = new Canvas(bi);
            canvas.drawColor(backcolor);                //Vorherige Zeichnung löschen
            canvas.drawLine(1, 1, xWidth-1, 1, scalepaint);    //Rand um Kurvenanzeige zeichnen
            canvas.drawLine(xWidth-1, 1, xWidth-1, dHeight - 1, scalepaint);
            canvas.drawLine(xWidth-1, dHeight - 1, 1, dHeight - 1, scalepaint);
            canvas.drawLine(1, dHeight - 1, 1, 1, scalepaint);
            canvas.drawText("Logik",100,100,scalepaint);
            I_L.setImageBitmap(bi);
        } else {
            I_L.setVisibility(View.INVISIBLE);
            aLP.height=1;
        }
        I_L.setLayoutParams(aLP);


    }

    private void ProtStreamFileSchreiben() {
        //Schreibt ein Messprotokoll-File, Filename wird automatisch gesetzt, da Eingabe eines Names evtl unpraktisch ist (z.B. bei Messung am fahrenden Fahrzeug)
        String filename, text;

        filename = Konfig_File.substring(0,Konfig_File.length()-4);       //Basis-Name aus Filename der Konfiguration nehmen (ohne ".kmk")
        filename = filename + "_";
        SimpleDateFormat sdf = new SimpleDateFormat("yyyyMMdd_HHmmss", Locale.getDefault());
        filename = filename + sdf.format(new Date()) + ".kmm";
        File writefile = new File(Environment.getExternalStorageDirectory(), WORKINGDIR + "/" +filename);
        try {
            int n;
            //File writefile = new File(filename);
            FileOutputStream filestream = new FileOutputStream(writefile, false);
            PrintStream printstream = new PrintStream(filestream);
            printstream.print("1.0:KMM - Messfile fuer den RemoteMessadapter" + "\n");
            printstream.print(Konfig_Text + ";" + Konfig_File + "\n");
            printstream.print(String.valueOf(Konfig_SampleRate) + "\n");
            printstream.print(String.valueOf(Konfig_BuffSize) + "\n");
            for (n=0;n<=5;n++) {
                if (Konfig_KanalAktiv[n]) {text="1;";} else {text="0;";}
                text = text + Konfig_KanalParam[n] + ";";
                text = text + Konfig_KanalEinheit[n] + ";";
                text = text + Konfig_KanalText[n];
                printstream.print(text + "\n");
            }
            if (Konfig_KanalAktiv[6]) {text="1";} else {text="0";}
            for (n=7;n<=9;n++) {
                if (Konfig_KanalAktiv[n]) {text=text+";1";} else {text=text+";0";}
            }
            printstream.print(text + "\n");
            for (n = 0; n < pstream.size(); n++) {
                printstream.print(pstream.get(n) +"\n");
            }
            filestream.close();
            Toast.makeText(getApplicationContext(), "File erstellt", Toast.LENGTH_SHORT).show();
        } catch (IOException e) {
            e.printStackTrace();
            Toast.makeText(MainActivity.this, "Fehler beim Schreiben", Toast.LENGTH_LONG).show();
        }
    }

    //region Bluetooth-Stuff
    private BluetoothAdapter keBTGetAdapter() {
        //Versucht einen Adapter zur Bluetooth-Hardware herzustellen
        final int REQUEST_ENABLE_BT = 1;
        BluetoothAdapter myBTAdapter;

        //Bluetooth-Adapter suchen
        myBTAdapter = BluetoothAdapter.getDefaultAdapter();
        if (myBTAdapter == null) {                   //Wird Bluetooth von dem Gerät unterstützt?
            Toast.makeText(this, "Kein Bluetooth vorhanden!", Toast.LENGTH_SHORT).show();
        } else {                                         //Wenn Bluetooth vorhanden aber nicht an ist: Nachfragen ob angeschaltet werden darf
            if (!myBTAdapter.isEnabled()) {
                Intent enableBtIntent = new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE);
                startActivityForResult(enableBtIntent, REQUEST_ENABLE_BT);
            }
        }
        return myBTAdapter;
    }

    private BluetoothDevice keBTGetDevice(BluetoothAdapter myBTAdapter, String DeviceName) {
        //Versucht Bluetooth-device mit dem entsprechenden Bluetooth-Namen zu finden und liefert entsprechendes device zurück
        Set<BluetoothDevice> pairedDevices;        //Liste der gepairten Devices
        String tempadd = null;
        BluetoothDevice myBTDevice;

        if (myBTAdapter != null) {
            if (myBTAdapter.isEnabled()) {
                // Suche gepairtes HC-05 Device
                pairedDevices = myBTAdapter.getBondedDevices();
                for (BluetoothDevice device : pairedDevices) {
                    if (device.getName().toString().equals(DeviceName)) {
                        tempadd = device.getAddress().toString();
                    }
                }
                final String address = tempadd;
                if (address == null) {        //Kein gepairtes HC-05 Device gefunden
                    Toast.makeText(getApplicationContext(), "Kein " + DeviceName + " gepaired", Toast.LENGTH_SHORT).show();
                    return null;
                } else {
                    myBTDevice = myBTAdapter.getRemoteDevice(address);
                }
                return myBTDevice;
            } else {
                return null;
            }
        } else {
            return null;
        }
    }

    private BluetoothSocket keBTConnectToSocket(BluetoothDevice myBTDevice) {
        //Stellt Verbindung zu Socket von device her
        final UUID BTMODULEUUID = UUID.fromString("00001101-0000-1000-8000-00805F9B34FB"); // "random" unique identifier
        BluetoothSocket myBTSocket;

        if (myBTDevice != null) {
            try {        //Erstelle BT Socket
                myBTSocket = myBTDevice.createRfcommSocketToServiceRecord(BTMODULEUUID);
            } catch (IOException e) {
                Toast.makeText(getApplicationContext(), "Fehler bei der Socket-Creation", Toast.LENGTH_SHORT).show();
                return null;
            }
            try {        //BT socket connection herstellen
                myBTSocket.connect();
            } catch (IOException e) {
                try {    //Geht connection schief: Socket schlieÃŸen
                    myBTSocket.close();
                    Toast.makeText(getApplicationContext(), "Fehler bei der Socket-Connection", Toast.LENGTH_SHORT).show();
                    return null;
                } catch (IOException e2) {
                    Toast.makeText(getApplicationContext(), "Fehler beim Schließen der Socke", Toast.LENGTH_SHORT).show();
                    return null;
                }
            }
            return myBTSocket;
        } else {
            return null;
        }
    }

    private OutputStream keBTGetOutputStream(BluetoothSocket myBTSocket) {
        //Erstellt einen Outputstream in den Bluetooth-Socket
        OutputStream outstr;

        if (myBTSocket != null) {
            try {
                outstr = myBTSocket.getOutputStream();
            } catch (IOException e) {
                Toast.makeText(getApplicationContext(), "Fehler bei der Erzeugung des Outputstream", Toast.LENGTH_SHORT).show();
                return null;
            }
            return outstr;
        } else {
            return null;
        }
    }

    private InputStream keBTGetInputStream(BluetoothSocket myBTSocket) {
        //Erstellt einen Inputstream aus dem Bluetooth-Socket
        InputStream instr;

        if (myBTSocket != null) {
            try {
                instr = myBTSocket.getInputStream();
            } catch (IOException e) {
                Toast.makeText(getApplicationContext(), "Fehler bei der Erzeugung des Inputstream", Toast.LENGTH_SHORT).show();
                return null;
            }
            return instr;
        } else {
            return null;
        }
    }

    private void keBTStartListening() {
        //Startet einen eigenen Thread, der durchgehend den InputStream ueberwacht, wenn mehr als RXbufsize neue Byte bereitliegen, wird onRXReceive aufgerufen
        final Handler handler = new Handler();
        keBTListening = true;

        keBTListenThread = new Thread(new Runnable() {
            public void run() {
                while (!Thread.currentThread().isInterrupted() && keBTListening) {
                    try {
                        int bytesAvailable = keBTInputStream.available();
                        if (bytesAvailable >= RXbufsize) {
                            byte[] RXpacket = new byte[RXbufsize];
                            keBTInputStream.read(RXpacket, 0, RXbufsize);
                            handler.post(new Runnable() {
                                //@Override
                                public void run() {
                                    onRXReceive(RXpacket);
                                }
                            });
                        }
                    } catch (IOException ex) {
                        keBTListening = false;
                    }
                }
            }
        });
        keBTListenThread.start();
    }

    private void keBTwriteByte(byte data) {
        //Schreibt ein Byte an Bluetooth
        try {
            keBTOutputStream.write(data);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void keBTflushReadBuffer() {
        //Leert den gesamten aktuell vorhandenen Read-Buffer
        try {
            int bytesAvailable = keBTInputStream.available();
            if (bytesAvailable > 0) {
                byte[] RXpacket = new byte[bytesAvailable];
                keBTInputStream.read(RXpacket, 0, bytesAvailable);
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
    //endregion Ende Bluetooth-Stuff

}